{% if router in ['fastapi', 'starlette'] %}
    {% include '_starlette_endpoints.py.jinja' %}
{% elif router == 'blacksheep' %}
    {% include '_blacksheep_endpoints.py.jinja' %}
{% elif router == 'litestar' %}
    {% include '_litestar_endpoints.py.jinja' %}
{% endif %}
